import tornado.web


class NoCacheRequestHandler(tornado.web.RequestHandler):
    """Base handler that adds no-cache headers to all responses."""

    def set_default_headers(self):
        """Set headers to prevent browser caching."""
        self.set_header(
            "Cache-Control", "no-store, no-cache, must-revalidate, max-age=0"
        )
        self.set_header("Pragma", "no-cache")
        self.set_header("Expires", "0")


class NoCacheStaticFileHandler(tornado.web.StaticFileHandler):
    """Static file handler that adds no-cache headers to all responses."""

    def set_default_headers(self):
        """Set headers to prevent browser caching."""
        self.set_header(
            "Cache-Control", "no-store, no-cache, must-revalidate, max-age=0"
        )
        self.set_header("Pragma", "no-cache")
        self.set_header("Expires", "0")



from RGBMatrixEmulator.cli.command import CLICommand

from RGBMatrixEmulator.internal.emulator_config import RGBMatrixEmulatorConfig


class ConfigCLICommand(CLICommand):
    def __init__(self, _arguments) -> None:
        pass

    def execute(self):
        RGBMatrixEmulatorConfig._dump_default_config()



import os
from RGBMatrixEmulator.adapters.browser_adapter.request_handlers import (
    NoCacheRequestHandler,
)


class SingleFileHandler(NoCacheRequestHandler):
    def initialize(self, file_path):
        self.file_path = file_path

    async def get(self):
        if not os.path.exists(self.file_path):
            self.set_status(404)
            return

        self.set_header("Content-Type", "image/x-icon")

        with open(self.file_path, "rb") as f:
            self.write(f.read())



import numpy as np
from PIL import Image, ImageEnhance
from RGBMatrixEmulator.graphics.color import Color


class Canvas:
    def __init__(self, options):
        self.options = options

        self.width = options.cols * options.chain_length
        self.height = options.rows * options.parallel

        # 3D numpy array -- rows (H), columns (W), 3-tuple RGB
        self.__pdims = (self.height, self.width, 3)

        self.display_adapter = options.display_adapter.get_instance(
            self.width, self.height, options
        )

        self.Clear()

        self.display_adapter.load_emulator_window()

    def Clear(self):
        self.__pixels = np.full(
            self.__pdims, self.__create_pixel(Color.BLACK()), dtype=np.uint8
        )

    def Fill(self, r, g, b):
        self.__pixels = np.full(
            self.__pdims, self.__create_pixel((r, g, b)), dtype=np.uint8
        )

    def SetPixel(self, x, y, r, g, b):
        if self.__pixel_out_of_bounds(x, y):
            return

        self.__pixels[int(y)][int(x)] = self.__create_pixel((r, g, b))

    def SetImage(self, image, offset_x=0, offset_y=0, *other):
        enhancer = ImageEnhance.Brightness(image)
        image = enhancer.enhance(self.brightness / 100.0)

        original = Image.fromarray(self.__pixels, "RGB")
        original.paste(image, (offset_x, offset_y))
        self.__pixels = np.copy(original)

    @property
    def brightness(self):
        return self.options.brightness

    @brightness.setter
    def brightness(self, value):
        if not isinstance(value, (int, float)):
            raise ValueError(f"brightness must be a numeric value, received '{value}'")
        elif value < 0 or value > 100:
            raise ValueError(
                f"brightness must be a number between 0 and 100, received '{value}'"
            )

        self.options.brightness = value

    def __create_pixel(self, pixel):
        return Color.adjust_brightness(tuple(pixel), self.brightness / 100.0)

    def __pixel_out_of_bounds(self, x, y):
        if x < 0 or x >= self.width:
            return True

        if y < 0 or y >= self.height:
            return True

        return False

    # These are delegated to the display adapter to handle specific implementation.
    def draw_to_screen(self):
        self.display_adapter.draw_to_screen(self.__pixels)

    def check_for_quit_event(self):
        self.display_adapter.check_for_quit_event()



from RGBMatrixEmulator.internal.emulator_config import RGBMatrixEmulatorConfig


class RGBMatrixOptions:
    def __init__(self):
        self.hardware_mapping = "EMULATED"
        self.rows = 32
        self.cols = 32
        self.chain_length = 1
        self.parallel = 1
        self.row_address_type = 0
        self.multiplexing = 0
        self.pwm_bits = 0
        self.brightness = 100
        self.pwm_lsb_nanoseconds = 130
        self.led_rgb_sequence = "RGB-EMULATED"
        self.show_refresh_rate = 0
        self.gpio_slowdown = None
        self.disable_hardware_pulsing = False

        emulator_config = RGBMatrixEmulatorConfig()

        self.display_adapter = emulator_config.display_adapter
        self.pixel_style = emulator_config.pixel_style
        self.pixel_glow = emulator_config.pixel_glow
        self.pixel_size = emulator_config.pixel_size
        self.pixel_outline = emulator_config.DEFAULT_CONFIG["pixel_outline"]
        self.pixel_outline = emulator_config.pixel_outline

        # Browser Adapter
        self.browser = emulator_config.browser
        self.emulator_title = emulator_config.emulator_title
        self.icon_path = emulator_config.icon_path

        # Pi5 Adapter
        self.pi5 = emulator_config.pi5

    def window_size(self):
        return (
            self.cols * self.pixel_size * self.chain_length,
            self.rows * self.pixel_size * self.parallel,
        )

    def window_size_str(self, pixel_text=""):
        width, height = self.window_size()

        return f"{width} x {height} {pixel_text}"



from RGBMatrixEmulator.adapters.browser_adapter.request_handlers.base import (
    NoCacheRequestHandler,
    NoCacheStaticFileHandler,
)
from RGBMatrixEmulator.adapters.browser_adapter.request_handlers.main import MainHandler
from RGBMatrixEmulator.adapters.browser_adapter.request_handlers.image import (
    ImageHandler,
)
from RGBMatrixEmulator.adapters.browser_adapter.request_handlers.image_web_socket import (
    ImageWebSocketHandler,
)
from RGBMatrixEmulator.adapters.browser_adapter.request_handlers.single_file import (
    SingleFileHandler,
)



ADAPTERS = {
    "browser": {
        "path": "RGBMatrixEmulator.adapters.browser_adapter.adapter",
        "class": "BrowserAdapter",
        "fallback": True,
    },
    "pygame": {
        "path": "RGBMatrixEmulator.adapters.pygame_adapter",
        "class": "PygameAdapter",
        "fallback": True,
    },
    "sixel": {
        "path": "RGBMatrixEmulator.adapters.sixel_adapter",
        "class": "SixelAdapter",
        "fallback": True,
    },
    "terminal": {
        "path": "RGBMatrixEmulator.adapters.terminal_adapter",
        "class": "TerminalAdapter",
        "fallback": True,
    },
    "tkinter": {
        "path": "RGBMatrixEmulator.adapters.tkinter_adapter",
        "class": "TkinterAdapter",
        "fallback": True,
    },
    "turtle": {
        "path": "RGBMatrixEmulator.adapters.turtle_adapter",
        "class": "TurtleAdapter",
        "fallback": True,
    },
    "raw": {
        "path": "RGBMatrixEmulator.adapters.raw_adapter",
        "class": "RawAdapter",
        "fallback": False,
    },
    "pi5": {
        "path": "RGBMatrixEmulator.adapters.pi5_adapter",
        "class": "Pi5Adapter",
        "fallback": False,
    },
}



from functools import wraps
from inspect import signature

from RGBMatrixEmulator.graphics.color import Color
from RGBMatrixEmulator.graphics.font import Font


def validate_color(func):
    """
    Decorator to validate that the 'color' argument is of type Color.
    If not, attempts to map to the rpi-rgb-led-matrix error message.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        # Map positional + keyword args to parameter names
        sig = signature(func)
        bound = sig.bind(*args, **kwargs)
        bound.apply_defaults()

        if "color" in bound.arguments:
            value = bound.arguments["color"]
            if not isinstance(value, Color):
                expected = f"{Color.__module__}.{Color.__qualname__}"
                actual = f"{type(value).__module__}.{type(value).__qualname__}"
                raise TypeError(
                    f"Argument 'color' has incorrect type "
                    f"(expected {expected}, got {actual})"
                )

        return func(*args, **kwargs)

    return wrapper


@validate_color
def DrawText(canvas, font, x, y, color, text):
    # Early return for empty string prevents bugs in bdfparser library
    # and makes good sense anyway
    if len(text) == 0:
        return

    # Support multiple spacings based on device width
    character_widths = [__actual_width(font, letter) for letter in text]
    first_char_width = character_widths[0]
    max_char_width = max(character_widths)
    total_width = sum(character_widths)

    # Offscreen to the left, adjust by first character width
    if x < 0:
        adjustment = abs(x + first_char_width) // first_char_width
        text = text[adjustment:]
        if adjustment:
            x += first_char_width * adjustment

    # Offscreen to the right, rough adjustment by max width
    if (total_width + x) > canvas.width:
        text = text[: ((canvas.width + 1) // max_char_width) + 2]

    # Draw the text!
    if len(text) != 0:
        # Ensure text doesn't get drawn as multiple lines
        linelimit = len(text) * (font.headers["fbbx"] + 1)

        text_map = font.bdf_font.draw(
            text, linelimit, missing=font.default_character
        ).todata(2)
        font_y_offset = -(font.headers["fbby"] + font.headers["fbbyoff"])

        for y2, row in enumerate(text_map):
            for x2, value in enumerate(row):
                if value == 1:
                    canvas.SetPixel(
                        x + x2,
                        y + y2 + font_y_offset,
                        color.red,
                        color.green,
                        color.blue,
                    )

    return total_width


@validate_color
def DrawLine(canvas, x1, y1, x2, y2, color):
    int_points = __coerce_int(x1, y1, x2, y2)
    rows, cols = __line(*int_points)

    for point in zip(rows, cols):
        canvas.SetPixel(*point, color.red, color.green, color.blue)


@validate_color
def DrawCircle(canvas, x, y, r, color):
    int_points = __coerce_int(x, y)
    rows, cols = __circle_perimeter(*int_points, r)

    for point in zip(rows, cols):
        canvas.SetPixel(*point, color.red, color.green, color.blue)


def __actual_width(font, letter):
    """
    Returns the actual width of the letter in the font. If the font doesn't contain a glyph for this letter, it falls back to
    the width of the default character (?) to prevent division by 0.
    """
    width = font.CharacterWidth(ord(letter))

    if width > 0:
        return width

    return font.CharacterWidth(font.default_character.cp())


def __coerce_int(*values):
    return [int(value) for value in values]


def __line(x1, y1, x2, y2):
    """
    Line drawing algorithm

    Extracted from scikit-image:
    https://github.com/scikit-image/scikit-image/blob/00177e14097237ef20ed3141ed454bc81b308f82/skimage/draw/_draw.pyx#L44
    """
    steep = 0
    r = x1
    c = y1
    dr = abs(x2 - x1)
    dc = abs(y2 - y1)

    rr = [0] * (max(dc, dr) + 1)
    cc = [0] * (max(dc, dr) + 1)

    if (y2 - c) > 0:
        sc = 1
    else:
        sc = -1
    if (x2 - r) > 0:
        sr = 1
    else:
        sr = -1
    if dr > dc:
        steep = 1
        c, r = r, c
        dc, dr = dr, dc
        sc, sr = sr, sc
    d = (2 * dr) - dc

    for i in range(dc):
        if steep:
            rr[i] = c
            cc[i] = r
        else:
            rr[i] = r
            cc[i] = c
        while d >= 0:
            r = r + sr
            d = d - (2 * dc)
        c = c + sc
        d = d + (2 * dr)

    rr[dc] = x2
    cc[dc] = y2

    return (rr, cc)


def __circle_perimeter(x, y, radius):
    """
    Bresenham circle algorithm

    Extracted from scikit-image
    https://github.com/scikit-image/scikit-image/blob/00177e14097237ef20ed3141ed454bc81b308f82/skimage/draw/_draw.pyx#L248
    """
    rr = list()
    cc = list()

    c = 0
    r = radius
    d = 3 - 2 * radius

    while r >= c:
        rr.extend([_ + x for _ in [r, -r, r, -r, c, -c, c, -c]])
        cc.extend([_ + y for _ in [c, c, -c, -c, r, r, -r, -r]])

        if d < 0:
            d += 4 * c + 6
        else:
            d += 4 * (c - r) + 10
            r -= 1
        c += 1

    return (rr, cc)



import time
from RGBMatrixEmulator.logger import Logger

DEFAULT_UPDATE_RATE = 0.5  # seconds


class FPSMonitor:
    def __init__(self, update_rate=DEFAULT_UPDATE_RATE):
        self.update_rate = update_rate
        self.n_frames = 0
        self.start_time = time.time()

    def tick(self):
        self.n_frames += 1
        now = time.time()
        elapsed = now - self.start_time

        if elapsed >= self.update_rate:
            fps = round(self.n_frames / elapsed, 2)

            Logger.debug(
                f"FPS: {fps} (received {self.n_frames} frames over {round(elapsed, 2)}s)"
            )

            self.n_frames = 0
            self.start_time = now



from RGBMatrixEmulator.adapters.browser_adapter.request_handlers import (
    NoCacheRequestHandler,
)


class ImageHandler(NoCacheRequestHandler):
    def get(self):
        self.set_header(
            "Content-type", "image/{}".format(self.adapter.image_format.lower())
        )
        self.write(self.adapter.image)

    def register_adapter(adapter):
        ImageHandler.adapter = adapter



import io, webbrowser
from pathlib import Path

from RGBMatrixEmulator.adapters.base import BaseAdapter
from RGBMatrixEmulator.internal.pixel_style import PixelStyle
from RGBMatrixEmulator.adapters.browser_adapter.server import Server
from RGBMatrixEmulator.logger import Logger


class BrowserAdapter(BaseAdapter):
    SUPPORTED_PIXEL_STYLES = [
        PixelStyle.SQUARE,
        PixelStyle.CIRCLE,
        PixelStyle.REAL,
    ]
    IMAGE_FORMATS = {"bmp": "BMP", "jpeg": "JPEG", "png": "PNG", "webp": "WebP"}

    def __init__(self, width, height, options):
        super().__init__(width, height, options)
        self.__server = None
        self.image = None
        self.image_ready = False
        self.default_image_format = "JPEG"

        image_format = options.browser.image_format
        if image_format.lower() in self.IMAGE_FORMATS:
            self.image_format = self.IMAGE_FORMATS[image_format.lower()]
        else:
            Logger.warning(
                "Invalid browser image format '{}', falling back to '{}'".format(
                    image_format, self.default_image_format
                )
            )
            self.image_format = self.IMAGE_FORMATS.get(
                self.default_image_format.lower()
            )

        # Default icon path is browser adapter assets
        self.default_icon_path = str(
            (Path(__file__).parent / "static" / "assets" / "icon.ico").resolve()
        )
        self._set_icon_path()

    def load_emulator_window(self):
        if self.loaded:
            return

        Logger.info(self.emulator_title)

        self.__server = Server(self)
        self.__server.run()

        self.loaded = True

        self.__open_browser()

    def draw_to_screen(self, pixels):
        image = self._get_masked_image(pixels)
        with io.BytesIO() as bytesIO:
            image.save(
                bytesIO,
                self.image_format,
                quality=self.options.browser.quality,
                optimize=True,
            )
            self.image = bytesIO.getvalue()

        self.image_ready = True

    def __open_browser(self):
        if self.options.browser.open_immediately:
            try:
                uri = f"http://localhost:{self.options.browser.port}"
                Logger.info(
                    f"Browser adapter configured to open immediately, opening new window/tab to {uri}"
                )
                webbrowser.open(uri)
            except Exception as e:
                Logger.exception("Failed to open a browser window")
                Logger.exception(e)



import json, os, pprint

from RGBMatrixEmulator.internal.pixel_style import PixelStyle
from RGBMatrixEmulator.internal.adapter_loader import AdapterLoader
from RGBMatrixEmulator.logger import Logger


class RGBMatrixEmulatorConfig:
    CONFIG_PATH = "emulator_config.json"

    DEFAULT_CONFIG = {
        "pixel_outline": 0,
        "pixel_size": 16,
        "pixel_style": "square",
        "pixel_glow": 6,
        "display_adapter": "browser",
        "allow_adapter_fallback": True,
        "icon_path": None,
        "emulator_title": None,
        "suppress_font_warnings": False,
        "browser": {
            "_comment": "For use with the browser adapter only.",
            "port": 8888,
            "target_fps": 60,
            "fps_display": False,
            "quality": 70,
            "image_border": True,
            "debug_text": False,
            "image_format": "JPEG",
            "open_immediately": False,
        },
        "pi5": {
            "_comment": "For use with the pi5 adapter only.",
            "pinout": "AdafruitMatrixBonnet",
            "n_addr_lines": 4,
            "rotation": "Normal",
            "n_planes": 10,
            "n_temporal_planes": 4,
            "n_lanes": 1,
            "led_rgb_sequence": "RGB",
        },
        "log_level": "info",
    }

    def __init__(self):
        self.config = self.__load_config()
        self.default_config = self.DEFAULT_CONFIG

        RGBMatrixEmulatorConfig.Utils.set_attributes(self)

        loader = AdapterLoader(requested_adapter=self.display_adapter.lower())
        self.display_adapter = loader.load(fallback=self.allow_adapter_fallback)

        self.__validate_pixel_style()
        self.__validate_pixel_glow()

        if self.suppress_font_warnings:
            import bdfparser

            bdfparser.warnings.simplefilter("ignore")

    def __load_config(self):
        if os.path.exists(self.CONFIG_PATH):
            with open(self.CONFIG_PATH) as f:
                config = json.load(f)

            return config

        Logger.info("Existing emulator config not found...")
        self._dump_default_config()

        return self.DEFAULT_CONFIG

    @classmethod
    def _dump_default_config(cls):
        Logger.info("Creating a new default emulator config.")
        with open(cls.CONFIG_PATH, "w") as f:
            json.dump(cls.DEFAULT_CONFIG, f, indent=4)
        Logger.info("Created emulator_config.json.")

    def __validate_pixel_style(self):
        requested_style = self.pixel_style.upper()
        self.pixel_style = PixelStyle.fetch(requested_style)

        if self.pixel_style.name == requested_style:
            if self.pixel_style not in self.display_adapter.SUPPORTED_PIXEL_STYLES:
                self.pixel_style = PixelStyle.DEFAULT
                Logger.warning(
                    """
"{}" pixel style option is not supported by adapter "{}". 
Supported pixel styles for this adapter are {}
                                                              
Defaulting to "{}"...
""".format(
                        self.pixel_style.lower(),
                        self.display_adapter.lower(),
                        ", ".join(
                            f'"{style.config_name}"'
                            for style in self.display_adapter.SUPPORTED_PIXEL_STYLES
                        ),
                        PixelStyle.DEFAULT.config_name,
                    )
                )
        else:
            Logger.warning(
                '"{}" pixel style option not recognized. Valid options are {}. Defaulting to "{}"...'.format(
                    self.pixel_style.lower(),
                    ", ".join(f'"{style.config_name}"' for style in list(PixelStyle)),
                    PixelStyle.DEFAULT.config_name,
                )
            )

    def __validate_pixel_glow(self):
        if not (isinstance(self.pixel_glow, int) and self.pixel_glow >= 0):
            Logger.warning(
                '"{}" pixel glow option not recognized. Valid options are integers >= 0. Defaulting to {}...'.format(
                    self.pixel_glow, self.DEFAULT_CONFIG.get("pixel_glow")
                )
            )

            self.pixel_glow = self.DEFAULT_CONFIG.get("pixel_glow")

    def __str__(self):
        return RGBMatrixEmulatorConfig.Utils.to_str(self)

    class ChildConfig:
        def __init__(self, config, default_config):
            self.config = config
            self.default_config = default_config

            RGBMatrixEmulatorConfig.Utils.set_attributes(self)

        def __str__(self):
            return RGBMatrixEmulatorConfig.Utils.to_str(self)

    class Utils:
        def to_str(obj):
            """
            Pretty prints the config object from dict.
            """
            printer = pprint.PrettyPrinter(sort_dicts=False)
            return "\n".join(
                [
                    obj.__repr__(),
                    printer.pformat(RGBMatrixEmulatorConfig.Utils.to_dict(obj)),
                ]
            )

        def to_dict(obj):
            """
            Recursively recreates the config dict from child config objects.
            """
            config = {}
            for key in obj.__dict__.keys():
                if key in ["config", "default_config"] or key[0] == "_":
                    continue

                value = obj.__dict__.get(key)

                if isinstance(value, RGBMatrixEmulatorConfig.ChildConfig):
                    value = RGBMatrixEmulatorConfig.Utils.to_dict(value)

                config[key] = value

            return config

        def set_attributes(obj):
            """
            Dynamically set attributes loaded into config and default config variables.

            Numbers, strings, and arrays are stored natively. Nested dicts are parsed into RGBMatrixEmulatorChildConfig objects recursively.
            """
            for key in obj.default_config.keys():
                if key in obj.config:
                    value = obj.config.get(key)
                    default = obj.default_config.get(key)
                else:
                    value = obj.default_config.get(key)
                    default = value

                    Logger.warning(
                        "Emulator config is missing key '{}', falling back to default '{}'. Consider adding this to your emulator config file.".format(
                            key, value
                        )
                    )

                RGBMatrixEmulatorConfig.Utils.set_attribute(obj, key, value, default)

        def set_attribute(obj, key, value, default):
            """
            Store the value as an attribute or delegate to the RGBMatrixEmulatorChildConfig to parse into a new node.
            """
            if isinstance(value, dict):
                obj.__setattr__(
                    key, RGBMatrixEmulatorConfig.ChildConfig(value, default)
                )
            else:
                obj.__setattr__(key, value)



import argparse, sys

from RGBMatrixEmulator.cli.config import ConfigCLICommand
from RGBMatrixEmulator.logger import Logger


class CLI:
    COMMANDS = {
        # Full commands
        "config": ConfigCLICommand,
        # Aliases
        "c": ConfigCLICommand,
    }

    @staticmethod
    def execute():
        parser = argparse.ArgumentParser(
            description="RGBMatrixEmulator command line interface."
        )
        subparsers = parser.add_subparsers(
            dest="command", required=True, help="Available commands"
        )

        # "config" command
        subparsers.add_parser(
            "config",
            aliases=["c"],
            help="Initialize and manipulate emulator_config.json files",
        )

        args = parser.parse_args()

        if args.command not in CLI.COMMANDS:
            Logger.critical(f"Error: Unknown command '{args.command}'")
            sys.exit(1)

        cmd = CLI.COMMANDS[args.command]
        cmd(args).execute()


def run_cli():
    CLI.execute()



import os
import sys

from RGBMatrixEmulator.adapters.base import BaseAdapter
from RGBMatrixEmulator.internal.pixel_style import PixelStyle


class TerminalAdapter(BaseAdapter):
    SUPPORTED_PIXEL_STYLES = [PixelStyle.SQUARE, PixelStyle.CIRCLE]
    SYMBOLS = {PixelStyle.SQUARE: "██", PixelStyle.CIRCLE: " ●"}

    def __init__(self, width, height, options):
        super().__init__(width, height, options)
        self.__symbol = self.SYMBOLS.get(self.options.pixel_style)

    def draw_to_screen(self, pixels):
        output = "\033[H\n"  # Move the cursor to the home position, add a little border
        for pixel_row in pixels:
            output += "  "  # Add a bit of border in case cursor causes line to wrap
            for pixel in pixel_row:
                output += "\033[38;2;{};{};{}m".format(
                    *pixel
                )  # Set the cell to the pixel color
                output += self.__symbol  # Draw the pixel
                output += "\033[37m"  # Reset the color

            output += " \n"

        sys.stdout.write(output)

    def load_emulator_window(self):
        os.system("cls||clear")
        os.system(
            "mode con: cols={} lines={}".format(self.width * 2 + 5, self.height + 3)
        )



from RGBMatrixEmulator.graphics import *
from RGBMatrixEmulator.emulation.matrix import RGBMatrix
from RGBMatrixEmulator.emulation.options import RGBMatrixOptions

from RGBMatrixEmulator.version import __version__



from RGBMatrixEmulator.adapters.browser_adapter.request_handlers import (
    NoCacheRequestHandler,
)


class MainHandler(NoCacheRequestHandler):
    def get(self):
        self.render("./../static/index.html", adapter=MainHandler.adapter)

    def register_adapter(adapter):
        MainHandler.adapter = adapter



from RGBMatrixEmulator.emulation.canvas import Canvas


class RGBMatrix:
    def __init__(self, options={}):
        self.options = options

        self.width = options.cols * options.chain_length
        self.height = options.rows * options.parallel

        self.canvas = None

    def CreateFrameCanvas(self):
        self.canvas = Canvas(options=self.options)

        return self.canvas

    # TODO: framerate_fraction support not yet implemented
    # https://github.com/hzeller/rpi-rgb-led-matrix/commit/f88355e46faafb9de7f1e59dd258ab36c7e7b097
    def SwapOnVSync(self, canvas, *other):
        canvas.check_for_quit_event()
        canvas.draw_to_screen()
        self.canvas = canvas

        return self.canvas

    def Clear(self):
        self.__sync_canvas()
        self.canvas.Clear()
        self.SwapOnVSync(self.canvas)

    def Fill(self, r, g, b):
        self.__sync_canvas()
        self.canvas.Fill(r, g, b)
        self.SwapOnVSync(self.canvas)

    def SetPixel(self, x, y, r, g, b):
        self.__sync_canvas()
        self.canvas.SetPixel(x, y, r, g, b)
        self.SwapOnVSync(self.canvas)

    def SetImage(self, image, offset_x=0, offset_y=0, *other):
        self.__sync_canvas()
        self.canvas.SetImage(image, offset_x, offset_y, *other)
        self.SwapOnVSync(self.canvas)

    def __sync_canvas(self):
        if not self.canvas:
            self.canvas = Canvas(options=self.options)

    @property
    def brightness(self):
        return self.options.brightness

    @brightness.setter
    def brightness(self, value):
        if not isinstance(value, (int, float)):
            raise ValueError(f"brightness must be a numeric value, received '{value}'")
        elif value < 0 or value > 100:
            raise ValueError(
                f"brightness must be a number between 0 and 100, received '{value}'"
            )

        self.options.brightness = value



import numpy as np
from pathlib import Path

from PIL import Image, ImageDraw
from RGBMatrixEmulator import version
from RGBMatrixEmulator.logger import Logger
from RGBMatrixEmulator.internal.pixel_style import PixelStyle


class BaseAdapter:
    SUPPORTED_PIXEL_STYLES = [PixelStyle.DEFAULT]
    INSTANCE = None

    DEFAULT_MASK_FN = "_draw_square_mask"
    MASK_FNS = {
        PixelStyle.SQUARE: "_draw_square_mask",
        PixelStyle.CIRCLE: "_draw_circle_mask",
        PixelStyle.REAL: "_draw_real_mask",
    }

    DEBUG_TEXT_TEMPLATE = (
        "RGBME v{} - {}x{} Matrix | {}x{} Chain | {}px per LED ({}) | {}"
    )

    ICON_FORMATS = ["PNG", "ICO", "JPEG"]
    ICON_MAX_SIZE = (512, 512)

    def __init__(self, width, height, options):
        self.width = width
        self.height = height
        self.options = options
        self.__black = Image.new("RGB", self.options.window_size(), "black")
        self.__mask = self.__draw_mask()
        self.loaded = False

        self.emulator_title = self.options.emulator_title or str(self)

        self.default_icon_path = (Path(__file__).parent / ".." / "icon.png").resolve()
        self._set_icon_path()

    @classmethod
    def get_instance(cls, *args, **kwargs):
        if cls.INSTANCE is None:
            instance = cls(*args, **kwargs)
            cls.INSTANCE = instance

        return cls.INSTANCE

    # This method is required for the pygame adapter but nothing else, so just skip it if not defined.
    def check_for_quit_event(self):
        pass

    #############################################################
    # These methods must be implemented by BaseAdapter subclasses
    #############################################################
    def load_emulator_window(self):
        """
        Initialize the external dependency as a graphics display.

        This method is fired when the emulated canvas is initialized.
        """
        raise NotImplementedError

    def draw_to_screen(self, _pixels):
        """
        Accepts a 2D array of pixels of size height x width.

        Implements drawing each pixel to the screen via the external dependency loaded in load_emulator_window.
        """
        raise NotImplementedError

    #############################################################
    # These methods implement pixel masks (styles)
    #############################################################
    def _get_masked_image(self, pixels):
        image = Image.fromarray(np.array(pixels, dtype=np.uint8), "RGB")
        image = image.resize(self.options.window_size(), Image.NEAREST)

        return Image.composite(image, self.__black, self.__mask)

    def __mask_fn(self, pixel_style):
        if pixel_style not in self.MASK_FNS:
            Logger.warning(
                f"Pixel style '{pixel_style.config_name}' mask function not found, defaulting to {self.DEFAULT_MASK_FN}..."
            )

        return getattr(self, self.MASK_FNS.get(pixel_style, self.DEFAULT_MASK_FN))

    def __draw_mask(self):
        mask = Image.new("L", self.options.window_size())

        draw_fn = self.__mask_fn(self.options.pixel_style)
        draw_fn(mask)

        return mask

    def _draw_circle_mask(self, mask):
        pixel_size = self.options.pixel_size
        width, height = self.options.window_size()

        drawer = ImageDraw.Draw(mask)

        for y in range(0, height, pixel_size):
            for x in range(0, width, pixel_size):
                drawer.ellipse(
                    (x, y, x + pixel_size - 1, y + pixel_size - 1),
                    fill=255,
                    outline=255,
                )

    def _draw_square_mask(self, mask):
        pixel_size = self.options.pixel_size
        width, height = self.options.window_size()

        drawer = ImageDraw.Draw(mask)

        for y in range(0, height, pixel_size):
            for x in range(0, width, pixel_size):
                drawer.rectangle(
                    (x, y, x + pixel_size, y + pixel_size),
                    fill=255,
                    outline=255,
                )

    def _draw_real_mask(self, mask):
        pixel_size = self.options.pixel_size
        width, height = self.options.window_size()
        pixel_glow = self.options.pixel_glow

        if pixel_glow == 0:
            # Short circuit to a faster draw routine
            return self._draw_circle_mask(mask)

        # Create two gradients.
        # The first is the LED with a gradient amount of 1 to antialias the result.
        # The second is the actual glow given the setting.
        gradient = self._gradient_add(
            self._generate_gradient(pixel_size, 1),
            self._generate_gradient(pixel_size, pixel_glow),
        )

        pixel = Image.fromarray(gradient.astype(np.uint8))

        # Paste the pixel into the mask at each point.
        for y in range(0, height, pixel_size):
            for x in range(0, width, pixel_size):
                mask.paste(pixel, (x, y), pixel)

    def _generate_gradient(self, sz, amt):
        """
        Generates a radial gradient of size sz with glow amt.

        The resulting array is normalized between [0, 255].
        """
        # Calculate our own radial gradient to use in the mask.
        # PIL.Image.radial_gradient() produces subpar results (LEDs look blocky)
        x = np.linspace(0, sz, sz)
        y = np.linspace(0, sz, sz)

        # Discrete points (x, y) between 0 and pixel_size
        X, Y = np.meshgrid(x, y)

        # Distance of point to center
        center = sz / 2
        D = np.sqrt((X - center) ** 2 + (Y - center) ** 2)

        # LED radius
        L = (sz - amt) / 2
        G = amt

        # Calculate the radial gradient glow, clip it between 0 and 1, scale by 255 (WHITE), and get the additive inverse
        gradient = 255 - np.clip((D - L) / G, 0, 1) * 255

        return gradient

    def _gradient_add(self, g1, g2):
        """
        Adds two arrays g1 and g2 with equal weight, and normalizes between [0, 255].
        Assumes the input arrays are normalized between [0, 255].

        Each pair of points is added with equal weight by first subtracting 128,
        which produces fewer graphical artifacts than summation by average of the two points.
        """
        return np.clip((g1 - 128) + (g2 - 128), 0, 255)

    def _set_icon_path(self):
        self.icon_path = self.default_icon_path
        custom_path = self.options.icon_path

        if not custom_path:
            return

        try:
            icon = Image.open(custom_path)

            if icon.format not in self.ICON_FORMATS:
                Logger.info(
                    f"Custom icon format '{icon.format}' is not in allowed formats ({self.ICON_FORMATS}). Using default icon instead."
                )
                return

            if (
                icon.width > self.ICON_MAX_SIZE[0]
                or icon.height > self.ICON_MAX_SIZE[1]
            ):
                Logger.info(
                    f"Icon of size '{icon.size}' is too large (max size is {self.ICON_MAX_SIZE}). Using default icon instead."
                )
                return
        except Exception as e:
            Logger.exception(
                "Encountered exception while loading custom icon. Using default icon instead."
            )
            Logger.exception(e)
            return

        self.icon_path = self.options.icon_path

    def __str__(self):
        return self.DEBUG_TEXT_TEMPLATE.format(
            version.__version__,
            self.options.cols,
            self.options.rows,
            self.options.chain_length,
            self.options.parallel,
            self.options.pixel_size,
            self.options.pixel_style.name,
            self.__class__.__name__,
        )



class Color:
    def __init__(self, r=0, g=0, b=0):
        # Simulate Cython uint8_t bounds checking
        for c in (r, g, b):
            if c > 255:
                raise OverflowError("value too large to convert to uint8_t")

            if c < 0:
                raise OverflowError("can't convert negative value to uint8_t")

        self.red = r
        self.green = g
        self.blue = b

    @classmethod
    def adjust_brightness(cls, pixel, alpha, to_int=False):
        if to_int:
            return tuple(int(channel) for channel in pixel)

        return tuple(channel * alpha for channel in pixel)

    @classmethod
    def to_hex(cls, pixel):
        return "#%02x%02x%02x" % pixel

    @classmethod
    def BLACK(cls):
        return (0, 0, 0)

    @classmethod
    def RED(cls):
        return (255, 0, 0)

    @classmethod
    def GREEN(cls):
        return (0, 255, 0)

    @classmethod
    def BLUE(cls):
        return (0, 0, 255)



from RGBMatrixEmulator.adapters.base import BaseAdapter

from PIL import Image

import numpy as np


class RawAdapter(BaseAdapter):
    MAX_FRAMES_STORED = 128
    DEFAULT_MAX_FRAME = -1  # Never halts

    def __init__(self, width, height, options):
        super().__init__(width, height, options)

        self._reset()

    def draw_to_screen(self, pixels):
        self.frames[self.frame] = pixels

        if self.frame - self.MAX_FRAMES_STORED in self.frames:
            del self.frames[self.frame - RawAdapter.MAX_FRAMES_STORED]

        self.frame += 1

        if self.halt_after > 0 and self.frame >= self.halt_after:
            self.halt_fn()

    def load_emulator_window(self):
        pass

    def _dump_screenshot(self, path):
        image = Image.fromarray(np.array(self._last_frame(), dtype="uint8"), "RGB")
        image.save(path)

    def _last_frame(self):
        return self.frames[self.frame - 1]

    def _reset(self):
        self.frames = {}
        self.frame = 0

        self.halt_after = RawAdapter.DEFAULT_MAX_FRAME
        self.halt_fn = lambda: 1 + 1



import importlib, sys, traceback

from enum import Enum, auto

from RGBMatrixEmulator.logger import Logger

from RGBMatrixEmulator.adapters import ADAPTERS


class LoadResult(Enum):
    SUCCESS = auto()
    FAILURE = auto()
    SKIPPED = auto()

    DEFAULT = SKIPPED


class AdapterLoader:

    def __init__(self, requested_adapter):
        self.requested_adapter = requested_adapter.lower()
        self.adapter_priority = [self.requested_adapter] + [
            adapter_name
            for adapter_name in list(ADAPTERS.keys())
            if adapter_name != self.requested_adapter
        ]
        self.adapter = None

        self._loaded = False
        self._adapters = {}

    def load(self, fallback=False):
        if self._loaded:
            return self.adapter

        exceptions = []

        for adapter_name in self.adapter_priority:
            result, adapter, tb = self._load(adapter_name, fallback=fallback)

            if result == LoadResult.SKIPPED:
                continue
            elif result == LoadResult.FAILURE:
                if tb is not None:
                    exceptions.append((adapter_name, tb))
            else:
                self._loaded = True
                self.adapter = adapter

                break

        if self._loaded:
            return self.adapter

        Logger.critical("All available display adapters failed to load!")

        for adapter_name, tb in exceptions:
            Logger.error(f"Adapter '{adapter_name}' failed with error:")
            Logger.error(tb)

        sys.exit(1)

    def _load(self, adapter_name, fallback=False):
        try:
            adapter_config = ADAPTERS[adapter_name]
            package_path = adapter_config.get("path")
            adapter_class = adapter_config.get("class")
            valid_fallback = adapter_config.get("fallback")

            if self.requested_adapter != adapter_name and not valid_fallback:
                return (LoadResult.SKIPPED, None, None)

            package = importlib.import_module(package_path)
            adapter = getattr(package, adapter_class)

            self._adapters[adapter_name] = adapter

            return (LoadResult.SUCCESS, adapter, None)
        except Exception as e:
            if fallback:
                return (LoadResult.FAILURE, None, traceback.format_exc())

            Logger.exception("Display adapter load failed with no fallback configured!")
            Logger.exception(e)

            sys.exit(1)



import os
import platform
import sys
import io
import libsixel as sixel

from PIL import Image, ImageDraw, ImageEnhance
from typing import List

from RGBMatrixEmulator.adapters.base import BaseAdapter
from RGBMatrixEmulator.internal.pixel_style import PixelStyle
from RGBMatrixEmulator.graphics.color import Color


class SixelAdapter(BaseAdapter):
    SUPPORTED_PIXEL_STYLES = [PixelStyle.SQUARE, PixelStyle.CIRCLE]

    def draw_to_screen(self, pixels):
        sixel = self.__encode_sixels(
            pixels, scale=self.options.pixel_size, outline=self.options.pixel_outline
        )
        output = f"\033[H\n{sixel}\n"
        sys.stdout.write(output)

    def load_emulator_window(self):
        os.system("cls||clear")
        if platform.system() == "Windows":
            os.system(
                "mode con: cols={} lines={}".format(self.width * 2 + 5, self.height + 3)
            )
        else:
            os.system(
                "stty columns {} rows {}".format(self.width * 2 + 5, self.height + 3)
            )

    def __enlarge_pixels(
        self, pixels: List[List[Color]], scale: int, outline: int
    ) -> Image.Image:
        outline_color = "#000000"
        w, h = len(pixels[0]), len(pixels)
        iw, ih = w * scale, h * scale
        i = Image.new("RGBA", (iw, ih))
        d = ImageDraw.Draw(i)
        for y, row in enumerate(pixels):
            for x, pixel in enumerate(row):
                rx, ry = x * scale, y * scale
                ex, ey = rx + (scale - 1), ry + (scale - 1)
                if self.options.pixel_style == "circle":
                    d.ellipse(
                        [(rx, ry), (ex, ey)],
                        fill=pixel,
                        outline=outline_color,
                        width=outline,
                    )
                else:
                    d.rectangle(
                        [(rx, ry), (ex, ey)],
                        fill=pixel,
                        outline=outline_color,
                        width=outline,
                    )

        brightness = ImageEnhance.Brightness(i)
        i = brightness.enhance(1.5)
        contrast = ImageEnhance.Contrast(i)
        i = contrast.enhance(0.8)

        return i

    def __encode_sixels(self, pixels: List[List[Color]], scale=4, outline=1) -> str:
        """Encodes given Image to a sixel string."""
        img = self.__enlarge_pixels(pixels, scale, outline)
        img_data = img.convert("RGB").tobytes()
        width = img.width
        height = img.height
        with io.BytesIO() as buf:
            output = sixel.sixel_output_new(
                lambda data, buffer: buffer.write(data), buf
            )
            dither = sixel.sixel_dither_new(256)
            sixel.sixel_dither_initialize(
                dither, img_data, width, height, sixel.SIXEL_PIXELFORMAT_RGB888
            )
            sixel.sixel_encode(img_data, width, height, 1, dither, output)

            encoded = buf.getvalue().decode("ascii")

            sixel.sixel_dither_unref(dither)
            sixel.sixel_output_unref(output)

        return encoded



import asyncio
import signal
import sys
import threading
import tornado.web
import tornado.ioloop

from pathlib import Path
from tornado.platform.asyncio import AnyThreadEventLoopPolicy

from RGBMatrixEmulator.adapters.browser_adapter.request_handlers import *
from RGBMatrixEmulator.logger import Logger

asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())


class Server:
    instance = None

    # Singleton class for the server.
    # No more than one webserver can be running at a given time.
    class __Singleton:
        def __init__(self, adapter):
            self.adapter = adapter
            self.io_loop = None
            self.listening = False

            MainHandler.register_adapter(self.adapter)
            ImageWebSocketHandler.register_adapter(self.adapter)
            ImageHandler.register_adapter(self.adapter)

            script_path = Path(__file__).resolve().parent
            asset_path = script_path / "static" / "assets"

            self.app = tornado.web.Application(
                [
                    (r"/websocket", ImageWebSocketHandler),
                    (r"/image", ImageHandler),
                    (r"/", MainHandler),
                    # Icons are served by a specific handler to allow custom icons at arbitrary filepaths
                    (
                        r"/assets/icon.ico",
                        SingleFileHandler,
                        {"file_path": adapter.icon_path},
                    ),
                    (
                        r"/assets/(.*)",
                        NoCacheStaticFileHandler,
                        {"path": asset_path, "default_filename": "client.js"},
                    ),
                ]
            )

            self.periodic = tornado.ioloop.PeriodicCallback(
                ImageWebSocketHandler.broadcast,
                1000 / adapter.options.browser.target_fps,
            )

            self.periodic.start()

    def __init__(self, adapter):
        if not Server.instance:
            Server.instance = Server.__Singleton(adapter)

    def run(self):
        if not self.instance.listening:
            Logger.info("Starting server...")

            self.instance.listening = True
            self.instance.app.listen(self.instance.adapter.options.browser.port)
            self.instance.io_loop = tornado.ioloop.IOLoop.current()
            thread = threading.Thread(
                target=self.instance.io_loop.start,
                name="RGBMEServerThread",
                daemon=True,
            )
            self.__initialize_interrupts()
            thread.start()

            Logger.info(
                "Server started and ready to accept requests on http://localhost:"
                + str(self.instance.adapter.options.browser.port)
                + "/"
            )

    def __initialize_interrupts(self):
        """
        Add custom signal handling to ensure webserver thread exits appropriately.

        Not thread-safe, signal handling must happen on the main thread.
        """
        if threading.current_thread() is threading.main_thread():
            signal.signal(signal.SIGINT, self.__kill)
            signal.signal(signal.SIGTERM, self.__kill)

    def __kill(self, *_args):
        self.instance.io_loop.add_callback(self.instance.io_loop.stop)
        sys.exit()



import os
import sys

# Try to suppress the pygame load warning if able.
try:
    os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "hide"
except Exception:
    pass

import pygame

from pygame.locals import QUIT
from RGBMatrixEmulator.adapters.base import BaseAdapter
from RGBMatrixEmulator.internal.pixel_style import PixelStyle
from RGBMatrixEmulator.logger import Logger


class PygameAdapter(BaseAdapter):
    SUPPORTED_PIXEL_STYLES = [
        PixelStyle.SQUARE,
        PixelStyle.CIRCLE,
        PixelStyle.REAL,
    ]

    def __init__(self, width, height, options):
        super().__init__(width, height, options)
        self.__surface = None

    def load_emulator_window(self):
        if self.loaded:
            return

        Logger.info(self.emulator_title)
        self.__surface = pygame.display.set_mode(self.options.window_size())
        pygame.init()

        self.__set_emulator_icon()
        pygame.display.set_caption(self.emulator_title)

        self.loaded = True

    def draw_to_screen(self, pixels):
        image = self._get_masked_image(pixels)
        pygame_surface = pygame.image.fromstring(
            image.tobytes(), self.options.window_size(), "RGB"
        )
        self.__surface.blit(pygame_surface, (0, 0))

        pygame.display.flip()

    def check_for_quit_event(self):
        # We don't have events, but this will keep the emulator from appearing as if it's not responding.
        # This also enables closing the window to kill the emulator
        for event in pygame.event.get():
            if event.type == QUIT:
                pygame.quit()
                sys.exit()

    def __set_emulator_icon(self):
        icon = pygame.image.load(self.icon_path)
        pygame.display.set_icon(icon)



from enum import Enum, auto


class PixelStyle(Enum):
    SQUARE = auto()
    CIRCLE = auto()
    REAL = auto()

    DEFAULT = SQUARE

    @property
    def config_name(self):
        return self.name.lower()

    @classmethod
    def fetch(cls, key, default=None):
        try:
            return cls[key]
        except KeyError:
            return default or cls.DEFAULT



from RGBMatrixEmulator.adapters.base import BaseAdapter
from RGBMatrixEmulator.logger import Logger

import sys
from pathlib import Path
import numpy as np

try:
    import adafruit_blinka_raspberry_pi5_piomatter as piomatter
except ImportError:
    piomatter = None


class Pi5Adapter(BaseAdapter):
    def __init__(self, width, height, options):
        super().__init__(width, height, options)
        self.matrix = None
        self.framebuffer = None
        self.pixel_swizzle = None

        self.__ensure_pi5_runtime()

    def load_emulator_window(self):
        config = self.options.pi5

        # Validate n_planes
        if config.n_planes > 10:
            config.n_planes = 10
        elif config.n_planes < 1:
            Logger.warning(
                f"n_planes must be at least 1. Got {config.n_planes}, setting to 1."
            )
            config.n_planes = 1

        # Validate n_temporal_planes
        valid_temporal = [0, 2, 4]
        if config.n_temporal_planes not in valid_temporal:
            closest = min(
                valid_temporal, key=lambda x: abs(x - config.n_temporal_planes)
            )
            Logger.warning(
                f"Invalid n_temporal_planes {config.n_temporal_planes}. "
                f"Snapping to closest valid value: {closest}."
            )
            config.n_temporal_planes = closest

        if config.n_temporal_planes > config.n_planes:
            # Find max valid value <= n_planes
            new_val = max(
                [v for v in valid_temporal if v <= config.n_planes], default=0
            )
            Logger.warning(
                f"n_temporal_planes ({config.n_temporal_planes}) cannot be greater than n_planes ({config.n_planes}). "
                f"Reducing to {new_val}."
            )
            config.n_temporal_planes = new_val

        if config.n_addr_lines == 5 and self.height < 64:
            Logger.critical(
                f"n_addr_lines=5 requires a display height of at least 64 pixels. Current height: {self.height}"
            )
            sys.exit(1)

        # Configure Geometry
        pinout_str = config.pinout

        # Handle LED Sequence (RGB vs BGR vs Other)
        # Check if BGR is requested and if there is a BGR pinout variant
        sequence = config.led_rgb_sequence.upper()

        if sequence == "BGR":
            # Attempt to find the BGR variant of the pinout
            # Typical naming convention: AdafruitMatrixBonnet -> AdafruitMatrixBonnetBGR
            bgr_pinout_str = f"{pinout_str}BGR"
            if hasattr(piomatter.Pinout, bgr_pinout_str):
                pinout_str = bgr_pinout_str
                # Handled by hardware/pinout, so we don't need to software swizzle
                sequence = "RGB"  # Treat as normal for rest of pipeline
            else:
                Logger.warning(
                    f"BGR Sequence requested but pinout '{bgr_pinout_str}' not found in piomatter.Pinout. Will attempt software pixel swap."
                )

        # Resolve Pinout Enum
        if hasattr(piomatter.Pinout, pinout_str):
            pinout = getattr(piomatter.Pinout, pinout_str)
        else:
            Logger.warning(
                f"Pinout '{pinout_str}' not found in piomatter.Pinout. Defaulting to AdafruitMatrixBonnet."
            )
            pinout = piomatter.Pinout.AdafruitMatrixBonnet

        # Setup Software Swizzle if needed
        # We need to map RGB (original) to Dest Sequence
        # Common cases: RGB (0,1,2), RBG (0,2,1), GRB (1,0,2), GBR (1,2,0), BRG (2,0,1), BGR (2,1,0)
        # Note: If we handled BGR via pinout, sequence is set to "RGB" above so we skip this.
        if sequence != "RGB":
            swizzle_map = {
                "RGB": [0, 1, 2],
                "RBG": [0, 2, 1],
                "GRB": [1, 0, 2],
                "GBR": [1, 2, 0],
                "BRG": [2, 0, 1],
                "BGR": [2, 1, 0],
            }

            if sequence in swizzle_map:
                self.pixel_swizzle = swizzle_map[sequence]
                Logger.info(
                    f"Using software pixel swizzle for sequence '{sequence}': {self.pixel_swizzle}"
                )
            else:
                Logger.warning(
                    f"Unknown LED sequence '{sequence}'. Display colors may be incorrect."
                )

        rotation_str = config.rotation
        if hasattr(piomatter.Orientation, rotation_str):
            rotation = getattr(piomatter.Orientation, rotation_str)
        else:
            Logger.warning(
                f"Rotation '{rotation_str}' not found in piomatter.Orientation. Defaulting to Normal."
            )
            rotation = piomatter.Orientation.Normal

        if "Active3" not in pinout_str and config.n_lanes > 2:
            Logger.warning(
                f"Pinout '{pinout_str}' does not support multiple lanes. "
                "Setting n_lanes to 2."
            )
            config.n_lanes = 2

        if config.n_lanes == 2 and "Active3" not in pinout_str:
            # Simple Geometry
            geometry = piomatter.Geometry(
                width=self.width,
                height=self.height,
                n_planes=config.n_planes,
                n_addr_lines=config.n_addr_lines,
                n_temporal_planes=config.n_temporal_planes,
                rotation=rotation,
            )
        else:
            if config.n_lanes < 2:
                Logger.warning(
                    f"Active3 pinouts require n_lanes >= 2. Got {config.n_lanes}, setting to 2."
                )
                config.n_lanes = 2

            pixelmap = None
            if config.n_lanes > 1 or "Active3" in pinout_str:
                try:
                    from adafruit_blinka_raspberry_pi5_piomatter.pixelmappers import (
                        simple_multilane_mapper,
                    )

                    pixelmap = simple_multilane_mapper(
                        self.width, self.height, config.n_addr_lines, config.n_lanes
                    )
                except ImportError:
                    Logger.warning(
                        "Could not import simple_multilane_mapper. Active3/Multilane support may fail."
                    )

            geometry = piomatter.Geometry(
                width=self.width,
                height=self.height,
                n_addr_lines=config.n_addr_lines,
                n_planes=config.n_planes,
                n_temporal_planes=config.n_temporal_planes,
                n_lanes=config.n_lanes,
                map=pixelmap,
            )

        self.framebuffer = np.zeros(
            shape=(geometry.height, geometry.width, 3), dtype=np.uint8
        )

        self.matrix = piomatter.PioMatter(
            colorspace=piomatter.Colorspace.RGB888Packed,
            pinout=pinout,
            framebuffer=self.framebuffer,
            geometry=geometry,
        )

        # Register cleanup to clear the screen on exit
        import atexit

        def cleanup():
            if self.framebuffer is not None and self.matrix is not None:
                self.framebuffer.fill(0)
                self.matrix.show()
                # Deinit if possible, though show() with 0s might be enough
                # self.matrix.deinit() # PioMatter object might not support deinit or it's automatic on GC

        atexit.register(cleanup)

        self.loaded = True

    def draw_to_screen(self, pixels):
        if not self.loaded:
            return

        # pixels is (height, width, 3)
        try:
            pixel_data = np.array(pixels, dtype=np.uint8)

            # Apply software swizzle if configured
            if self.pixel_swizzle:
                pixel_data = pixel_data[:, :, self.pixel_swizzle]

            np.copyto(self.framebuffer, pixel_data)
            self.matrix.show()
        except Exception as e:
            # Throttle logging?
            pass

    def __ensure_pi5_runtime(self):
        """
        Validates that the code is running on a Raspberry Pi 5
        and has the necessary dependencies installed.
        """
        is_pi5 = False
        try:
            model_path = Path("/proc/device-tree/model")
            if model_path.exists():
                model_name = model_path.read_text().strip()
                if "Raspberry Pi 5" in model_name:
                    is_pi5 = True
        except Exception:
            pass

        if not is_pi5:
            Logger.critical(
                "This module is designed exclusively for the Raspberry Pi 5."
            )
            sys.exit(1)

        if piomatter is None:
            Logger.critical(
                "Pi5 adapter cannot load due to missing dependencies for Raspberry Pi 5.\n"
                "Please install dependencies using the [pi5] feature option:\n"
                "    pip install RGBMatrixEmulator[pi5]"
            )
            sys.exit(1)



import tornado.websocket

from RGBMatrixEmulator.logger import Logger
from RGBMatrixEmulator.adapters.browser_adapter.fps import FPSMonitor

FPS_UPDATE_RATE = 10  # seconds
FPS = FPSMonitor(FPS_UPDATE_RATE)


class ImageWebSocketHandler(tornado.websocket.WebSocketHandler):
    clients = set()
    adapter = None

    @classmethod
    def broadcast(cls):
        if not ImageWebSocketHandler.adapter.image_ready:
            return

        if not ImageWebSocketHandler.adapter.image:
            Logger.warning(
                "No image received from {}!".format(
                    ImageWebSocketHandler.adapter.__class__.__name__
                )
            )
            return

        io_loop = tornado.ioloop.IOLoop.current()

        for client in list(cls.clients):
            io_loop.add_callback(
                client.write_message, ImageWebSocketHandler.adapter.image, binary=True
            )

        FPS.tick()

    def check_origin(self, _origin):
        # Allow access from every origin
        return True

    def open(self):
        ImageWebSocketHandler.clients.add(self)
        Logger.info("WebSocket opened from: " + self.request.remote_ip)

    def on_message(self, _message):
        if not ImageWebSocketHandler.adapter.image:
            Logger.warning(
                "No image received from {}!".format(
                    ImageWebSocketHandler.adapter.__class__.__name__
                )
            )
            return

        image_bytes = ImageWebSocketHandler.adapter.image
        self.write_message(image_bytes, binary=True)

    def on_close(self):
        ImageWebSocketHandler.clients.remove(self)

    def register_adapter(adapter):
        ImageWebSocketHandler.adapter = adapter
